# -*- coding: utf-8 -*-
"""
Functional algorithms for the design and purchase cost estimation of
vacuum systems.

References
----------
.. [1] Seider, W. D.; Lewin, D. R.; Seader, J. D.; Widagdo, S.; Gani, R.;
    Ng, M. K. Cost Accounting and Capital Cost Estimation.
    In Product and Process Design Principles; Wiley, 2017; pp 426–485.
.. [2] Amazon. Robinair (15115) VacuMaster Single Stage Vacuum Pump - Single-Stage, 1.5 CFM. 
    https://www.amazon.com/Robinair-15115-VacuMaster-Single-Vacuum/dp/B005CO9FDW?ref_=ast_sto_dp.
    Accessed on 09/28/2023.
.. [3] Amazon. Robinair (15300) VacuMaster Economy Vacuum Pump - 2-Stage, 3 CFM.
    https://www.amazon.com/Robinair-15300-VacuMaster-Economy-Vacuum/dp/B000O1E5UQ?ref_=ast_sto_dp
    Accessed on 09/28/2023.

"""
from numpy import log as ln
from . import mechanical as mch
from math import ceil
from biosteam.utils import checkbounds
from biosteam.exceptions import DesignError
from numba import njit
import biosteam as bst

__all__ = ('compute_vacuum_system_power_and_cost',)


# %% Data

# System types of vacuum systems
# Volumetric flowrate ranges, (cfm) and lower limit of suction (torr)
# Rotary vane pumps based on ref. [2], [3]
_steamjet_ejectors = {
    'One stage':               ((10, 1000000), 100),
    'Two stage':               ((10, 1000000),  15),
    'Three stage':             ((10, 1000000),   2)}
_liquid_ring = {
    'One stage water seal':    (( 3,   18000),  50),
    'Two stage water seal':    (( 3,   18000),  25),
    'Oil seal':                (( 3,   18000),  10)}
_dry_vacuum = {
    'Three stage rotary lobe': ((60,     240), 1.5),
    'Three stage claw':        ((60,     270), 0.3),
    'Screw compressor':        ((50,    1400), 0.1)}
_rotary_vane = {
    'One stage':               ((0,     1.51), 0.115),
    'Two stage':               ((1.5,   3.01), 0.035)}

_default_vacuum_systems = {'Liquid-ring pump': _liquid_ring,
                           'Steam-jet ejector': _steamjet_ejectors,
                           'Dry-vacuum pump': _dry_vacuum,
                           'Rotary-vane pump': _rotary_vane}

_default_rotary_vane_work_cost = {
    'One stage': (1/5 * 0.7457, 127*1.08), # hp to kW; 2023 USD (including tax & shipping)
    'Two stage': (1/3 * 0.7457, 248*1.08)
}

_air_density = 1.2041 # kg/m3 dry air

# %% Calculate vacuum system requirements

def compute_vacuum_system_power_and_cost(
        F_mass, F_vol, P_suction, vessel_volume,
        vacuum_system_preference=None):
    """
    Return a dictionary of vacuum system requirements.
    
    Parameters
    ----------
    F_mass : float
        Vapor mass flow rate entering vacuum system from vessel in kg/hr (not including inleakage).
    F_vol : float
        Vapor volumetric flow rate entering vacuum system from vessel in m3/hr (not including inleakage).
    P_suction : float
        Suction pressure in Pa
    vessel_volume : float
        Vacuum volume in m3
    vacuum_system_preference : 'Liquid-ring pump', 'Steam-jet ejector', or 'Dry-vacuum pump'
        Name(s) of preferred vacuum systems
    
    """
    P_suction *= 7.5006e-3 # to torr
    if vessel_volume:
        vessel_volume *= 35.315 # to ft3
        F_mass_air = calculate_air_inleakage(vessel_volume, P_suction) # lb/hr
        F_vol_air = 0.26697*F_mass_air/_air_density # cfm
    else:
        F_vol_air = F_mass_air = 0
    F_vol_cfm = 0.5886*F_vol + F_vol_air
    # if F_vol_cfm < 3.01:
        # factor = 3.01/F_vol_cfm
        # F_vol_cfm = 3.01
    # else:
    #     factor = 1
    factor = 1
    F_mass_kgph = (F_mass + 0.4536*F_mass_air)*factor # kg/hr
    F_mass_lbph = 2.205 * F_mass_kgph
    vacuum_systems = get_prefered_vacuum_systems(vacuum_system_preference)
    name, grade, N = select_vacuum_system(vacuum_systems, F_vol_cfm, P_suction, bool(vacuum_system_preference))
    base_cost = calculate_vacuum_cost(name, grade, F_mass_lbph, F_vol_cfm, P_suction)
    cost = bst.CE / 567.  * base_cost
    if name == 'Steam-jet ejector':
        for agent in bst.HeatUtility.heating_agents:
            if agent.P > 689475.: break # 100 psig
        steam = 0.41631 * F_mass_kgph # [kmol/hr] 7.5 weight steam/ weight gas
        work = 0.
        has_condenser = grade != 'One stage'
    elif name == 'Rotary-vane pump':
        has_condenser = False
        agent = None
        steam = 0.
        N = 1
        work = _default_rotary_vane_work_cost[grade][0]
    else:
        has_condenser = False
        agent = None
        steam = 0.
        work = calculate_vacuum_power(F_mass_kgph, P_suction)
    return {'Work': work, 
            'Cost': N * cost, 
            'Name': f"{name}, {grade.lower()}",
            'In parallel': N,
            'Condenser': has_condenser,
            'Steam flow rate': steam, 
            'Heating agent': agent}


# %% Supporting functions

def get_prefered_vacuum_systems(preference):
    defaults = _default_vacuum_systems
    if preference is None:
        return defaults
    else:
        if isinstance(preference, str):
            if preference not in defaults:
                raise ValueError(f"preference must be in the following list: {list(defaults)}")
            preference = (preference, *defaults)
        else:
            for name in preference:
                if name not in defaults:
                    raise ValueError(f"preference must be in the following list: {list(defaults)}")
        return {name: defaults[name] for name in preference}

def get_available_vacuum_systems(F_vol_cfm, P_suction):
    """
    Return available vacuum type and grade
    
    Parameters
    ----------
    F_vol_cfm : float
        Vapor volumetric flow rate entering vacuum system from vessel in cfm (including inleakage).
    P_suction : float
        Suction pressure in Torr
    """
    types = []
    for vacuumtype, vacuum_sys in _default_vacuum_systems.items():
        for grade, flowrange_minsuction in vacuum_sys.items():
            flowrange, minsuction = flowrange_minsuction
            if checkbounds(F_vol_cfm, flowrange) and P_suction > minsuction:
                types.append((vacuumtype, grade))
    return types

def select_vacuum_system(vacuum_systems, F_vol_cfm, P_suction, ignore_F_lb=False):
    """
    Return a heuristic vacuum type and grade
    
    Parameters
    ----------
    F_vol_cfm : float
        Vapor volumetric flow rate entering vacuum system from vessel in cfm (including inleakage).
    P_suction : float
        Suction pressure in Torr
    """
    for name, vacuum_sys in vacuum_systems.items():
        for grade, flowrange_minsuction in vacuum_sys.items():
            flowrange, minsuction = flowrange_minsuction
            if ignore_F_lb:
                if F_vol_cfm < flowrange[-1] and P_suction > minsuction:
                    return (name, grade, 1)
            elif checkbounds(F_vol_cfm, flowrange) and P_suction > minsuction:
                    return (name, grade, 1)
    for name, vacuum_sys in vacuum_systems.items(): # Flow rate too large
        for grade, flowrange_minsuction in vacuum_sys.items():
            flowrange, minsuction = flowrange_minsuction
            if P_suction > minsuction:
                N = ceil(F_vol_cfm / flowrange[-1])
                return (name, grade, N)
    raise DesignError(f'no vacuum system available at current flow ({F_vol_cfm:.2f} cfm) and suction pressure ({P_suction:.2f} Torr)')

@njit(cache=True)
def calculate_heuristic_air_inleakage(V, P):
    """
    Return air in-leakage in kg/hr through a heuristic calculation.
    
    Parameters
    ----------
    V : float
        Vacuum volume in m3
    P : float
        Suction pressure in Pa
    
    """
    if P > 11999.013: k = 0.2
    elif P > 4132.993: k = 0.15
    elif P > 2799.77: k = 0.10
    elif P > 133.322: k = 0.051
    else: raise ValueError('cannot calculate air inleakage at pressures lower than 133.32 Pascal')
    return k*V**0.667

@njit(cache=True)
def calculate_air_inleakage(V, P):
    """
    Return air in-leakage in kg/hr.
    
    Parameters
    ----------
    V : float
        Vacuum volume in ft3
    P : float
        Suction pressure in Torr
    """
    lnP = ln(P) 
    return 5. + (0.0298 + 0.03088*lnP - 5.733e-4*lnP*lnP)*V**0.66

def calculate_vacuum_power(F_mass,  P_suction):
    """
    Return vacuum power (after accounting for motor efficiency) in kW.
    
    Parameters
    ----------
    F_mass : float
        Total mass flow rate entering vacuum system in kg/hr (including inleakage).
    P_suction : float
        Suction pressure in Torr
    
    """
    SF = F_mass/P_suction # Size factor
    if SF < 0.2: SF = 0.2
    elif SF > 16.: SF = 16.
    Pb = 21.4*SF**0.924 # Break power assuming Liquid-ring (NASH)
    mu_m = mch.motor_efficiency(Pb)
    return Pb/mu_m

def calculate_vacuum_cost(vacuum_sys, grade, F_mass_lbph, F_vol_cfm, P_suction):
    # Costing for different vacuums ASSUMING THAT ALL ARE CARBON STEEL!!!!
    if vacuum_sys ==  'Steam-jet ejector':
        S = F_mass_lbph/P_suction
        Cp = 1915*S**0.41
        if grade == 'One stage':
            Cs = 1
        elif grade == 'Two stage':
            Cs = 1.8 * 1.6 # 2 stage + 1 condenser
        elif grade == 'Three stage':
            Cs = 2.1 * 2.3 # 3 stage + 2 condenser
        Cost = Cp*Cs
    elif vacuum_sys == 'Liquid-ring pump':
        S = F_vol_cfm
        Cp = 8250*S**0.37
        if grade == 'One stage water seal':
            Cs = 1
        elif grade == 'Two stage water seal':
            Cs = 1.8
        elif grade == 'Oil seal':
            Cs = 2.1
        Cost = Cp*Cs
    elif vacuum_sys == 'Dry-vacuum pump':
        S = F_vol_cfm
        if grade == 'Three stage rotary lobe':
            Cp = 8075*S**0.41
        elif grade == 'Three stage claw':
            Cp = 9785*S**0.36
        elif grade == 'Screw compressor':
            Cp = 10875*S**0.38
        Cost = Cp
    elif vacuum_sys == 'Rotary-vane pump':
        Cost = _default_rotary_vane_work_cost[grade][1] / 708. * 567.   # !!! 708 is the 2021 CEPCI, need to be updated to 2023 CEPCI
    return Cost
    

    
    


# -*- coding: utf-8 -*-
"""
"""
from dataclasses import dataclass
from thermosteam.utils import AbstractMethod
from colorpalette import Color
import biosteam as bst

__all__ = ('ProcessModel', 'ScenarioComparison')

def copy(scenario, **kwargs):
    for i in scenario.__slots__:
        if i not in kwargs: kwargs[i] = getattr(scenario, i)
    return scenario.__class__(**kwargs)
    
def scenario_info(scenario, add_metadata):
    slots = scenario.__slots__
    if add_metadata: 
        grey = Color(fg='#878787')
        metadata = scenario.metadata
    arguments = []
    for i in slots:
        j = getattr(scenario, i)
        if isinstance(j, (bool, str)):
            arg = f"{i}={j!r},"
        else:
            try:
                arg = f"{i}={j:.3g},"
            except:
                arg = f"{i}={j},"
        if add_metadata and i in metadata:
            comment = '# ' + metadata[i]
            separate = len(arg + comment) > 50
            comment = grey(comment)
            if separate:
                arguments.append(comment)
                arguments.append(arg)
                continue
            else:
                arg += ' ' + comment
        arguments.append(
            arg
        )
    if arguments:
        clsname = type(scenario).__name__
        N_spaces = len(clsname) +1
        if N_spaces > 4:
            arguments = '\n    '.join(arguments)
            return (
                f"{clsname}("
                f"\n    {arguments}\n"
                ")"
            
            )
        else:
            spaces = N_spaces * ' '
            arguments = f'\n{spaces}'.join(arguments)
            return (
                f"{clsname}({arguments})"
            
            )
    else:
        return (
            f"{type(scenario).__name__}()"
        )
    
def display_scenario(scenario, metadata=True):
    print(scenario_info(scenario, metadata))
    
def iterate_scenario_data(scenario):
    for i in scenario.__slots__:
        yield (i, getattr(scenario, i))

def scenario_comparison(left, right):
    return ScenarioComparison(left, right)

@dataclass(
    init=True, repr=True, eq=True, 
    unsafe_hash=False, frozen=True,
    match_args=True, slots=True,
)
class ScenarioComparison:
    left: object
    right: object

    def show(self):
        parameters = ',\n'.join(['left=' + scenario_info(self.left), 'right=' + scenario_info(self.right)])
        parameters = parameters.replace('\n', '\n    ')
        return print(
            f'{type(self).__name__}(\n'
            f'    {parameters}\n'
            f')'
        )
    _ipython_display_ = show
        

class ProcessModel:
    """
    ProcessModel is an abstract class which has missing (i.e., "abstract") 
    attributes and methods. The user must define a `Scenario` dataclass which
    determines all inputs to the process model. Additionally, the user must 
    define the `as_scenario`, `create_thermo`, `create_system`, and 
    `create_model` methods for the process model to initialize its key components. 
    
    It may help to look at how ProcessModel objects are created (approximately):
    
    .. code-block:: python
    
        def __new__(cls, simulate=None, scenario=None, **kwargs):
            if scenario is None:
                self.scenario = cls.Scenario(**kwargs)
            else:
                # The Scenario object can be initialized through the `as_scenario` class method.
                self.scenario = cls.as_scenario(scenario)
            
            # No need to recreate a process model for repeated scenarios.
            if scenario in cls.cache: return cls.cache[scenario]
            self = super().__new__()
            
            # The thermodynamic property package is given by the `create_thermo` method.
            self.load_thermo(self.create_thermo())
            
            # If no system is returned by the `create_system` method, a new system is created from flowsheet units.
            self.flowsheet = bst.Flowsheet()
            system = self.create_system()
            if system is None: system = self.flowsheet.create_system()
            
            # This saves the system as self.system and all units/streams as attributes by ID.
            # For example, Stream('feedstock') will be stored as self.feestock.
            self.load_system(system) 
            
            # A Model object is loaded from the `create_model` method.
            # The model will be stored as self.model and all parameters and metrics as attributes by function name.
            # For example: 
            #
            # @model.metric
            # def MSP(): return self.tea.solve_price(self.product)
            #
            # ^ This becomes self.MSP.
            self.load_model(self.create_model())
            
            if simulate: self.system.simulate()
            self.cache[scenario] = self
            return self
    
    """
    #: **class-attribute** Class which defines arguments to the process model using
    #: the layout of a python dataclass: https://docs.python.org/3/library/dataclasses.html
    Scenario: type
    
    #: This method allows the process model to default the scenario.
    #: It should return a Scenario object.
    default_scenario = AbstractMethod
    
    #: **class method** This method allows the process model to interpret objects 
    #: (e.g., strings, numbers) as a Scenario.
    as_scenario = AbstractMethod
    
    #: This method should return a chemicals or thermo object.
    #: BioSTEAM will automatically set it as the thermodynmic property package.
    create_thermo = AbstractMethod
    
    #: This method should create unit operations and connect them.
    #: It can return a system object, optionally. Otherwise, BioSTEAM will 
    #: take care of creating the system from the units and saves 
    #: it as the self.system attribute.
    #: All streams and unit operations are also saved as attributes by their ID.
    create_system = AbstractMethod
    
    #: This method should return a model object. 
    #: The model will be saved as a self.model attribute. 
    #: All pareameters and metrics of the model object will also be saved as 
    #: attributes by their function names.
    create_model = AbstractMethod
    
    @classmethod
    def scenario_hook(cls, scenario, kwargs):
        if scenario is None:
            if cls.default_scenario:
                scenario = cls.default_scenario()
            else:
                try:
                    return cls.Scenario(**kwargs)
                except:
                    raise NotImplementedError('missing class method `default_scenario`')
        if not isinstance(scenario, cls.Scenario):
            if cls.as_scenario:
                scenario = cls.as_scenario(scenario)
            else:
                raise NotImplementedError('missing class method `as_scenario`')
        if kwargs: scenario = scenario.copy(**kwargs)
        return scenario
    
    def __init_subclass__(cls):
        cls.cache = {}
        if '__new__' in cls.__dict__: return
        if not hasattr(cls, 'Scenario'):
            raise NotImplementedError(
                "ProcessModel sublcass missing a 'Scenario' attribute"
            )
        if 'Scenario' in cls.__dict__:
            Scenario = cls.Scenario
            Scenario.metadata = metadata = {}
            for i, j in tuple(Scenario.__dict__.items()):
                if i.startswith('__'): continue
                if isinstance(j, str):
                    if j[0] == '#':
                        delattr(Scenario, i)
                        metadata[i] = j.lstrip('# ')
                if isinstance(j, tuple) and len(j) == 2 and isinstance(j[-1], str):
                    value, j = j
                    if j[0] == '#':
                        setattr(Scenario, i, value)
                        metadata[i] = j.lstrip('# ')
                    
            # TODO: get defaults and separate units of measure
            cls.Scenario = Scenario = dataclass(cls.Scenario,
                init=True, repr=True, eq=True, unsafe_hash=False, frozen=True,
                match_args=True, slots=True,
            )
            Scenario.items = iterate_scenario_data
            Scenario._ipython_display_ = Scenario.show = display_scenario
            Scenario.copy = copy
            Scenario.__sub__ = scenario_comparison
    
    def __new__(cls, *, simulate=True, scenario=None, **kwargs):
        scenario = cls.scenario_hook(scenario, kwargs)
        if scenario in cls.cache: return cls.cache[scenario]
        self = super().__new__(cls)
        self.scenario = scenario
        self.flowsheet = bst.Flowsheet(repr(self))
        thermo = self.create_thermo()
        if thermo is NotImplemented:
            raise NotImplementedError(f"{cls.__name__!r} is missing the `create_thermo` method")
        self.load_thermo(thermo)
        bst.main_flowsheet.set_flowsheet(self.flowsheet)
        unit_registry = self.flowsheet.unit
        unit_registry.open_context_level()
        system = self.create_system()
        if system is NotImplemented:
            raise NotImplementedError(f"{cls.__name__!r} is missing the `create_system` method")
        else:
            units = unit_registry.close_context_level()
            if system is None: system = bst.System.from_units(units=units)
            self.load_system(system)
        model = self.create_model()
        if model is NotImplemented:
            raise NotImplementedError(f"{cls.__name__!r} is missing the `create_model` method")
        elif model is None:
            raise RuntimeError('`create_model` must return a biosteam.Model object')
        self.load_model(model)
        if simulate: system.simulate()
        cls.cache[scenario] = self
        return self
    
    def load_thermo(self, thermo):
        bst.settings.set_thermo(thermo)
        thermo = bst.settings.get_thermo()
        self.chemicals = thermo.chemicals
        self.thermo = thermo
    
    def baseline(self):
        sample = self.model.get_baseline_sample()
        return sample, self.model(sample)
    
    def load_system(self, system):
        self.system = system
        self.flowsheet = flowsheet = system.flowsheet
        self.__dict__.update(flowsheet.to_dict())
            
    def load_model(self, model):
        self.model = model
        for i in model.parameters:
            setattr(self, i.setter.__name__, i)
            if i.baseline is not None: i.setter(i.baseline)
        for i in model.metrics:
            setattr(self, i.getter.__name__, i)
            
    @property
    def parameters(self):
        return self.model._parameters
        
    @property
    def metrics(self):
        return self.model._metrics
            
    def __repr__(self):
        scenario = self.scenario
        scenario_name = type(scenario).__name__
        process_name = type(self).__name__
        N = len(scenario_name)
        return process_name + repr(self.scenario)[N:]
    
    def show(self, metadata=True):
        """Print representation of process model."""
        scenario = self.scenario
        scenario_name = type(scenario).__name__
        process_name = type(self).__name__
        N = len(scenario_name)
        info = scenario_info(scenario, metadata)
        print(process_name + info[N:])
    
    _ipython_display_ = show
    